"""Write Word Vectors to a file.

We provide the main :py:func:`~word_vectors.write.write` function that can write
to various vector serialization formats based on the passed :py:attr:`~word_vecotrs.FileType`.
There are also several convenience functions for writing specific formats.
"""

import struct
from operator import itemgetter
from typing import Union, IO, TextIO, BinaryIO, Optional, Iterable
from file_or_name import file_or_name
from word_vectors import Vocab, Vectors, FileType, DENSE_MAGIC_NUMBER
from word_vectors.utils import find_max, padded_bytes, to_vocab


def write(
    wf: Union[str, IO],
    vocab: Union[Vocab, Iterable[str]],
    vectors: Vectors,
    file_type: FileType,
    max_len: Optional[int] = None,
):
    """Write word vectors to a file.

    This function dispatches to on of the following word vector format writers based
    on the file of ``file_type``.

    - :py:func:`~word_vectors.write.write_glove`
    - :py:func:`~word_vectors.write.write_w2v_text`
    - :py:func:`~word_vectors.write.write_w2v`
    - :py:func:`~word_vectors.write.write_dense`

    Args:
        wf: The file we are writing to.
        vocab: The vocab mapping words -> ints.
        vectors: The vectors as a ``np.ndarray``.
        file_type: The format to use when writing the vectors to disk.
        max_len: The maximum length of a word in vocab. Only used when writing Dense vectors.

    Raises:
        ValueError: If the an unsupported file type is passed
    """
    if file_type is FileType.GLOVE:
        write_glove(wf, vocab, vectors)
    elif file_type is FileType.W2V:
        write_w2v(wf, vocab, vectors)
    elif file_type is FileType.W2V_TEXT or file_type is FileType.FASTTEXT or file_type is FileType.NUMBERBATCH:
        write_w2v_text(wf, vocab, vectors)
    elif file_type is FileType.DENSE:
        write_dense(wf, vocab, vectors, max_len)
    else:
        raise ValueError(f"FileType not understood, got: {file_type}")


@file_or_name(wf="w")
def write_glove(wf: Union[str, TextIO], vocab: Union[Vocab, Iterable[str]], vectors: Vectors):
    """Write vectors to a glove file.

    See :py:func:`word_vectors.read.read_glove` for a description of the file format and
    examples of common pre-trained embeddings that use this format.

    Args:
        wf: The file we are writing to
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
    """
    vocab = to_vocab(vocab) if not isinstance(vocab, dict) else vocab
    for word, idx in sorted(vocab.items(), key=itemgetter(1)):
        wf.write(" ".join([word, *map(str, vectors[idx])]) + "\n")


@file_or_name(wf="w")
def write_w2v_text(wf: Union[str, TextIO], vocab: Union[Vocab, Iterable[str]], vectors: Vectors):
    """Write vectors in the word2vec format in a text file.

    See :py:func:`word_vectors.read.read_w2v_text` for a description of the file format and
    examples of common pre-trained embeddings that use this format.

    Args:
        wf: The file we are writing to
        vocab: The vocab of words -> ints
        vectors: The vectors we are writing
    """
    wf.write(f"{len(vocab)} {vectors.shape[1]}\n")
    write_glove(wf, vocab, vectors)


@file_or_name(wf="wb")
def write_w2v(wf: Union[str, BinaryIO], vocab: Union[Vocab, Iterable[str]], vectors: Vectors):
    """Write vectors to the word2vec format as a binary file.

    See :py:func:`word_vectors.read.read_w2v` for a description of the file format and
    examples of common pre-trained embeddings that use this format.

    Args:
        wf: The file we are writing to
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
    """
    vocab = to_vocab(vocab) if not isinstance(vocab, dict) else vocab
    wf.write(f"{len(vocab)} {vectors.shape[1]}\n".encode("utf-8"))
    for word, idx in sorted(vocab.items(), key=itemgetter(1)):
        wf.write((word + " ").encode("utf-8") + vectors[idx].tobytes())


@file_or_name(wf="wb")
def write_dense(
    wf: Union[str, BinaryIO], vocab: Union[Vocab, Iterable[str]], vectors: Vectors, max_len: Optional[int] = None
):
    """Write vectors to a dense file.

    See :py:func:`word_vectors.read.read_dense` for a description of the file format.

    Args:
        wf: The file we are writing to.
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
        max_len: The longest length of the words as (``utf-8``) bytes.
    """
    vocab = to_vocab(vocab) if not isinstance(vocab, dict) else vocab
    max_len = find_max(vocab) if max_len is None else max_len
    for val in (DENSE_MAGIC_NUMBER, len(vocab), vectors.shape[1], max_len):
        wf.write(struct.pack("<Q", val))
    for word, idx in sorted(vocab.items(), key=itemgetter(1)):
        word_len = len(word.encode("utf-8"))
        word = padded_bytes(word, max_len)
        wf.write(struct.pack("<I", word_len))
        wf.write(word)
        wf.write(vectors[idx].tobytes())



"""Convert between word vector formats.

We provide the main :py:func:`~word_vectors.convert.convert` function for converting
between arbitrary formats based on the passed :py:attr:`~word_vectors.FileType` (or
by sniffing the input file with :py:func:`~word_vectors.read.sniff` when not provided)
as well as several convenience function for converting between different pairs of formats.
"""

from typing import Union, TextIO, BinaryIO, Optional
from word_vectors import FileType
from word_vectors.read import read
from word_vectors.write import write
from word_vectors.utils import find_max, create_output_path


# We don't know what mode to open the file in (text for things like Glove while
# binary for things like Word2Vec or Dense) we can't use the `@file_or_name`
# decorator directly but all the functions we call use that so we can handle
# all the file formats.
def convert(
    f: Union[str, TextIO, BinaryIO],
    output: Optional[str] = None,
    output_file_type: FileType = FileType.DENSE,
    input_file_type: Optional[FileType] = None,
):
    """Convert vectors from one format to another.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
        output_file_type: The vector serialization format to use when
            writing out the vectors.
        input_file_type: An explicit vector format to use when reading.
    """
    w, wv = read(f, input_file_type)
    len_ = find_max(w.keys())
    output = create_output_path(f, output_file_type) if output is None else output
    write(output, w, wv, output_file_type, len_)


def w2v_to_dense(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert binary Word2Vec formatted vectors to the Dense format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.DENSE, FileType.W2V)


def glove_to_dense(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert GloVe formatted vectors to the Dense format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.Dense, FileType.GLOVE)


def w2v_text_to_dense(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert text Word2Vec formatted vectors to the Dense format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.Dense, FileType.W2V_TEXT)


def w2v_to_w2v_text(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert binary Word2Vec formatted vectors to the Binary Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V_TEXT, FileType.W2V)


def w2v_to_glove(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert binary Word2Vec formatted vectors to the GloVe format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.GLOVE, FileType.W2V)


def w2v_text_to_glove(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert text Word2Vec formatted vectors to the Glove format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.GLOVE, FileType.W2V_TEXT)


def w2v_text_to_w2v(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert text Word2Vec formatted vectors to the binary Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V, FileType.W2V_TEXT)


def glove_to_w2v(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert GloVe formatted vectors to the binary Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V, FileType.GLOVE)


def glove_to_w2v_text(f: Union[str, TextIO], output: Optional[str] = None):
    """Convert GloVe formatted vectors to the text Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V_TEXT, FileType.GLOVE)


def dense_to_w2v(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert Dense formatted vectors to the binary Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V, FileType.DENSE)


def dense_to_w2v_text(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert Dense formatted vectors to the text Word2Vec format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    convert(f, output, FileType.W2V_TEXT, FileType.DENSE)


def dense_to_glove(f: Union[str, BinaryIO], output: Optional[str] = None):
    """Convert Dense formatted vectors to the GloVe format.

    Args:
        f: The file to read from.
        output: The name for the output file. If not provided we use the
            input file name with a modified extension.
    """
    covert(f, output, FileType.GLOVE, FileType.DENSE)



"""Read, Write, and Convert between different word vector serialization formats."""

__version__ = "3.0.0"

from typing import Dict, Tuple
from enum import Enum
import numpy as np


#: A mapping of word to integer index. This index is used pull the this words
#: vector from the matrix of word vectors.
Vocab = Dict[str, int]
#: The actual word vectors. These are always of rank 2 and have the shape ``[vocab size, vector size]``
Vectors = np.ndarray


class FileType(Enum):
    """An Enumeration of the Word Vector file types supported."""

    #: The format used by Glove. See :py:func:`~word_vectors.read.read_glove` for a
    #: description of file format and common pre-trained embeddings that use this format.
    GLOVE = "glove"
    #: The text format introduced by Word2Vec. See :py:func:`~word_vectors.read.read_w2v_text`
    #: for a description of the file format and common pre-trained embeddings that use this format.
    W2V_TEXT = "w2v-text"
    #: The binary format used by Word2Vec and pre-trained GoogleNews vectors. See
    #: :py:func:`~word_vectors.read.read_w2v` for a description of the file format and common
    #: pre-trained embeddings that use this format.
    W2V = "w2v"
    #: Our new Dense file format. See :py:func:`~word_vectors.read.read_dense` for a description of the file format.
    DENSE = "dense"
    #: The file format used to distribute FastText vectors, it is just the word2vec text format.
    #: See :py:func:`~word_vectors.read.read_w2v_text` for a description of the file format.
    FASTTEXT = "w2v-text"
    #: The file format used to distribute Numberbatch vectors, it is just the word2vec text format.
    #: See :py:func:`~word_vectors.read.read_w2v_text` for a description of the file format.
    NUMBERBATCH = "w2v-text"

    @classmethod
    def from_string(cls, value: str) -> "FileType":
        """Convert a string into the Enum value.

        Args:
            value: The string specifying the file type.

        Returns:
            The Enum value parsed from the string.

        Raises:
            ValueError: If the string wasn't able to be parsed into
                an Enum value.
        """
        value = value.lower()
        if value == "glove":
            return cls.GLOVE
        if value == "w2v_text" or value == "w2v-text":
            return cls.W2V_TEXT
        if value == "w2v":
            return cls.W2V
        if value == "dense":
            return cls.DENSE
        if value == "numberbatch":
            return cls.NUMBERBATCH
        if value in ("fasttext", "fast-text", "fast_text"):
            return cls.FASTTEXT
        return ValueError(f"Unable to understand file type, got: {value}")

    def __str__(self) -> str:
        """When calling ``str`` on an enum member output a value suitable for filenames"""
        return self.value


INT_SIZE = 4  #: The size of an int32 in bytes used when reading binary files.
FLOAT_SIZE = 4  #: The size of a float32 in bytes when reading a binary file.
LONG_SIZE = 8  #: The size of an int64 in bytes when reading binary files.
DENSE_HEADER = 4  #: The number of elements in the Dense format header.
DENSE_MAGIC_NUMBER = 2283  #: A magic number used to identify a Dense format file.


import word_vectors.read as read_module
import word_vectors.write as write_module
import word_vectors.convert as convert_module
from word_vectors.read import (
    read,
    read_with_vocab,
    read_w2v,
    read_w2v_with_vocab,
    read_w2v_text,
    read_w2v_with_vocab,
    read_glove,
    read_glove_with_vocab,
    read_dense,
    read_dense_with_vocab,
    verify_dense,
)
from word_vectors.convert import (
    convert,
    w2v_to_dense,
    w2v_to_glove,
    w2v_to_w2v_text,
    glove_to_w2v,
    glove_to_w2v_text,
    glove_to_dense,
    w2v_text_to_w2v,
    w2v_text_to_glove,
    w2v_text_to_dense,
    dense_to_glove,
    dense_to_w2v,
    dense_to_w2v_text,
)
from word_vectors.write import write, write_w2v, write_w2v_text, write_glove, write_dense



"""Utilities for working with word vector I/O."""

import os
import pathlib
from contextlib import contextmanager
from typing import Tuple, Iterable, Union, BinaryIO, IO, Callable
import numpy as np
from file_or_name import file_or_name
from word_vectors import Vocab, FileType


# The characters we define as "non-binary" when guessing if a file is binary.
ASCII_CHARACTERS = b"".join(map(lambda x: bytes((x,)), range(32, 127))) + b"\n\r\t\f\b"


def find_space(buf: bytes, offset: int) -> Tuple[str, int]:
    """Find the first space starting from offset and return word that spans the spaces and the new offset.

    Args:
        buf: The bytes buffer we are looking for a space in.
        offset: Where in the buffer we start looking.

    Returns:
        A (word, offset) tuple where word is the text (decoded from ``utf-8``) starting at
        the original offset until the first space. Offset is index of the location just
        after the space we just found.
    """
    i = offset + 1
    while buf[i : i + 1] != b" ":
        i += 1
    word = buf[offset:i].decode("utf-8")
    return word, i + 1


def find_max(words: Iterable[str]) -> int:
    """Get the max length of words (as bytes).

    Note:
        This finds the length in (``utf-8``) bytes, this could be different than the max length
        of the word as returned by ``len`` because ``len`` is run on the string objects which
        might encode to more bytes (in ``utf-8``) for example an emoji is often a single character
        but as bytes it could be a few.

    Args:
        words: The series of words.

    Returns:
        The length of the longest word.
    """
    return max(map(len, map(lambda x: x.encode("utf-8"), words)))


@file_or_name(f="rb")
def is_binary(
    f: Union[str, BinaryIO], block_size: int = 512, ratio: float = 0.30, text_characters: bytes = ASCII_CHARACTERS
) -> bool:
    """Guess if a file is binary or not.

    This is based on the implementation from `here`_

    .. _here: https://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python

    Args:
        f: The file we are testing.
        block_size: The amount of the file to read in for checking.
        ratio: How many non-ascii characters before we assume it is binary.
        text_characters: Characters that we define as text characters, the ratio of these characters
            to others is used to determine if the file was binary or not.

    Returns:
        True if the file is binary, False otherwise
    """
    # Because we are operating on an open file object we need to reset where we read from in case
    # people are going to start reading from it right away.
    with bookmark(f):
        block = f.read(block_size)
    # If there are null bytes then it must be binary
    if b"\x00" in block:
        return True
    # We are defining an empty file as a text file.
    elif not block:
        return False

    # Delete all the characters from `text_characters` to leave only the non_text ones
    non_text = block.translate(None, text_characters)
    # If there are more than ratio non-text characters we are a binary file.
    return len(non_text) / len(block) > ratio


@contextmanager
def bookmark(f: IO):
    """Bookmark where we are in a file so we can return.

    This is a context manager that lets us save our spot in an open file,
    to some operations on that file, and then return to the original stop.

    This is very useful for things like sniffing a file. If the file is
    already open and you read in some bytes to estimate the format you need
    to remember to reset to the start or else you will get wrong results.
    This context manager automates this. ::

        f.tell()
        >>> 120
        with bookmark(f):
            _ = f.read(1024)
            print(f.tell())
        >>> 1144
        f.tell()
        >>> 120

    Args:
        f: The file we are bookmarking.
    """
    start = f.tell()
    yield
    f.seek(start)


def padded_bytes(word: str, max_len: int) -> bytes:
    """Pad a word out so the byte representation is max_len.

    Args:
        word: The word we are padding out and converting to binary
        max_len: How far to pad the string

    Returns:
        The word as bytes and extended.
    """
    byte_words = word.encode("utf-8")
    return byte_words + b" " * (max_len - len(byte_words))


def to_vocab(words: Iterable[str]) -> Vocab:
    """Convert a series of words to a vocab mapping strings to ints.

    Args:
        words: The words in the vocab

    Returns:
        The Vocabulary
    """
    return {w: i for i, w in enumerate(words)}


def create_output_path(path: Union[str, IO, pathlib.PurePath], file_type: FileType) -> str:
    """Create the output path by stripping the extension and added a new one based on the vector format.

    Args:
        path: The path to the input file.
        file_type: The vector format we are converting to.

    Returns:
        The new output path with an extension determined by the file type.
    """
    if isinstance(path, (str, pathlib.PurePath)):
        path = str(path)
    else:
        path = path.name
    base, _ = os.path.splitext(path)
    return f"{base}.{file_type}"


def uniform_initializer(unif: float) -> Callable[[int], np.ndarray]:
    """Create a vector initialization function that takes a vector size as input.

    Args:
        unif: The bounds that the new vector will be initialized within

    Returns:
        A function that returns a uniformly random vector between ``-unif`` and ``unif``,
    """

    def _unif_initializer(vector_size: int) -> np.ndarray:
        return np.random.rand(-unif, unif, size=(vector_size,))

    return _unif_initializer



"""Read word vectors from a file.

We provide a main :py:func:`~word_vectors.read.read` function for reading vectors
from a file. The serialization format can be explicitly provided with by passing a
:py:attr:`~word_vectors.FileType` or automatically inferred using
:py:func:`~word_vectors.read.sniff`. There are also several provided convenience
functions for reading from specific formats.
"""

import re
import os
import mmap
import struct
import logging
import pathlib
from typing import Tuple, Union, IO, TextIO, BinaryIO, Optional, Iterator, Callable
import numpy as np
from file_or_name import file_or_name
from word_vectors import INT_SIZE, LONG_SIZE, FLOAT_SIZE, DENSE_HEADER, Vocab, Vectors, FileType, DENSE_MAGIC_NUMBER
from word_vectors.utils import find_space, is_binary, bookmark, uniform_initializer


GLOVE_TEXT = re.compile(r"^[^ ]+? (-?\d+?\.\d+? )+", re.MULTILINE)
GLOVE_BIN = re.compile(br"^[^ ]+? (-?\d+?\.\d+? )+", re.MULTILINE)
W2V_TEXT = re.compile(r"^\d+ \d+$", re.MULTILINE)
W2V_BIN = re.compile(br"^\d+ \d+$", re.MULTILINE)

LOGGER = logging.getLogger("word_vectors")


# We don't know what mode to open the file in (text for things like Glove while
# binary for things like Word2Vec or Dense) we can't use the `@file_or_name`
# decorator directly but all the functions we call use that so we can handle
# all the file formats.
def read(f: Union[str, TextIO, BinaryIO], file_type: Optional[FileType] = None) -> Tuple[Vocab, Vectors]:
    """Read vectors from a file.

    This function can dispatch to one of the following word vector format readers:

    - :py:func:`~word_vectors.read.read_glove`
    - :py:func:`~word_vectors.read.read_w2v_text`
    - :py:func:`~word_vectors.read.read_w2v`
    - :py:func:`~word_vectors.read.read_dense`

    Check the documentation of a specific reader to see a description of the file
    format as well as common pre-trained vectors that ship with this format.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Note:
        Without a specified file type this function uses :py:func:`word_vectors.read.sniff`
        to determine the word vector format and dispatches to the appropriate reader.

        I haven't seen a sniffing failure but if your file type can't be determined you
        can pass the ``file_type`` explicitly or call the specific reading function yourself.

    Args:
        f: The file to read from.
        file_type: The vector file format. If ``None`` the file is sniffed to determine
            format.

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if file_type is None:
        file_type = sniff(f)
        LOGGER.info("Sniffed word vector as type %s", file_type)
    if file_type is FileType.GLOVE:
        line_reader = read_glove_lines
    elif file_type is FileType.W2V_TEXT or file_type is FileType.FASTTEXT or file_type is FileType.NUMBERBATCH:
        line_reader = read_w2v_text_lines
    elif file_type is FileType.W2V:
        line_reader = read_w2v_lines
    elif file_type is FileType.DENSE:
        line_reader = read_dense_lines
    else:
        raise ValueError(f"Unknown vector format, got: {line_reader}")
    return _read(f, line_reader)


def read_with_vocab(
    f: Union[str, IO],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray] = uniform_initializer(0.25),
    keep_extra: bool = False,
    file_type: Optional[FileType] = None,
) -> Tuple[Vocab, Vectors]:
    """Read vectors from a file subject user provided vocabulary constraints.

    This function can dispatch to one of the following word vector format readers:

    - :py:func:`~word_vectors.read.read_glove_with_vocab`
    - :py:func:`~word_vectors.read.read_w2v_text_with_vocab`
    - :py:func:`~word_vectors.read.read_w2v_with_vocab`
    - :py:func:`~word_vectors.read.read_dense_with_vocab`

    Check the documentation of a specific reader to see a description of the file
    format as well as common pre-trained vectors that ship with this format.

    When provided a vocabulary this function will not reorder it. If you pass in that
    the word ``dog`` is index ``12`` then in the resulting vocabulary it will still
    be index ``12``.

    When collecting extra vocabulary (words that are in the pre-trained embeddings but
    not in the user vocab) these will all be at the end of the vocabulary. Again the
    indices of user provided words will not change.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Note:
        Without a specified file type this function uses :py:func:`word_vectors.read.sniff`
        to determine the word vector format and dispatches to the appropriate reader.

        I haven't seen a sniffing failure but if your file type can't be determined you
        can pass the ``file_type`` explicitly or call the specific reading function yourself.

    Args:
        f: The file to read from.
        user_vocab: A specific vocabulary the user wants to extract form the pre-trained
            embeddings.
        initializer: A function that takes the vector size and generates a new vector.
            this is used to generate a representation for a word in the user vocab that
            is not in the pre-train embeddings.
        keep_extra: Should you also include vectors that are in the pre-trained embedding
            but not in the user provided vocab?
        file_type: The vector file format. If ``None`` the file is sniffed to determine
            format.

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if file_type is None:
        file_type = sniff(f)
        LOGGER.info("Sniffed word vector as type %s", file_type)
    if file_type is FileType.GLOVE:
        line_reader = read_glove_lines
    elif file_type is FileType.W2V_TEXT or file_type is FileType.FASTTEXT or file_type is FileType.NUMBERBATCH:
        line_reader = read_w2v_text_lines
    elif file_type is FileType.W2V:
        line_reader = read_w2v_lines
    elif file_type is FileType.DENSE:
        line_reader = read_dense_lines
    else:
        raise ValueError(f"Unknown vector format, got: {line_reader}")
    if keep_extra:
        return _read_with_vocab_extra(f, line_reader, user_vocab, initializer)
    return _read_with_vocab(f, line_reader, user_vocab, initializer)


def _read(
    f: Union[str, IO], line_reader: Callable[[Union[str, IO]], Iterator[Tuple[str, np.ndarray]]]
) -> Tuple[Vocab, Vectors]:
    words = {}
    vectors = []
    for word, vector in line_reader(f):
        if word not in words:
            words[word] = len(words)
            vectors.append(vector)
    return words, np.vstack(vectors)


def _read_with_vocab(
    f: Union[str, IO],
    line_reader: Callable[[Union[str, IO]], Iterator[Tuple[str, np.ndarray]]],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray],
) -> Tuple[Vocab, Vectors]:
    vectors = [None for _ in range(len(user_vocab))]
    for word, vector in line_reader(f):
        vector_size = len(vector)
        if word in user_vocab:
            idx = user_vocab[word]
            if vectors[idx] is None:
                vectors[idx] = vector
    for i, vector in enumerate(vectors):
        if vector is None:
            vectors[i] = initializer(vector_size)
    return user_vocab, np.vstack(vectors)


def _read_with_vocab_extra(
    f: Union[str, IO],
    line_reader: Callable[[Union[str, IO]], Iterator[Tuple[str, np.ndarray]]],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray],
) -> Tuple[Vocab, Vectors]:
    user_vocab_size = len(user_vocab)
    vectors = [None for _ in range(len(user_vocab))]
    extra_words = {}
    extra_vectors = []
    for word, vector in line_reader(f):
        vector_size = len(vector)
        if word in user_vocab:
            idx = user_vocab[word]
            if vectors[idx] is None:
                vectors[idx] = vector
        else:
            if word not in extra_words:
                extra_words[word] = len(extra_words)
                extra_vectors.append(vector)
    for i, vector in enumerate(vectors):
        if vector is None:
            vectors[i] = initializer(vector_size)
    for word, idx in extra_words.items():
        user_vocab[word] = user_vocab_size + idx
    vectors.extend(extra_vectors)
    return user_vocab, np.vstack(vectors)


@file_or_name
def read_glove_lines(f: Union[str, TextIO]) -> Iterator[Tuple[str, np.ndarray]]:
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        for line in iter(m.readline, b""):
            line = line.decode("utf-8")
            line = line.rstrip("\n")
            word, *vector = line.split(" ")
            yield word, np.asarray(vector, dtype=np.float32)


@file_or_name
def read_w2v_text_lines(f: Union[str, TextIO]) -> Iterator[Tuple[str, np.ndarray]]:
    # Because the ``mmap`` call starts at the beginning of the file and the offset
    # needs to be a multiple of ``ALLOCATIONGRANULARITY`` we can't start from the offset
    # that an ``f.readline()`` would give us. This means we can't just advance by one line
    # and then call read_glove so I had to duplicate code here :/
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        _ = m.readline()
        for line in iter(m.readline, b""):
            line = line.decode("utf-8")
            line = line.rstrip("\n")
            word, *vector = line.split(" ")
            yield word, np.asarray(vector, dtype=np.float32)


@file_or_name(f="rb")
def read_w2v_lines(f: Union[str, BinaryIO]) -> Iterator[Tuple[str, np.ndarray]]:
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        header = m.readline()
        offset = m.tell()
        vocab, dim = map(int, header.decode("utf-8").split())
        size = FLOAT_SIZE * dim
        for _ in range(vocab):
            word, offset = find_space(m, offset)
            raw = m[offset : offset + size]
            vector = np.frombuffer(raw, dtype=np.float32)
            yield word, vector
            offset = offset + size


@file_or_name(f="rb")
def read_dense_lines(f: Union[str, BinaryIO]) -> Iterator[Tuple[str, np.ndarray]]:
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        offset = LONG_SIZE * DENSE_HEADER
        vocab, dim, max_length = read_dense_header(m[:offset])
        size = FLOAT_SIZE * dim
        line_length = max_length + size + INT_SIZE
        for i in range(vocab):
            start = offset + i * (line_length)
            line = m[start : start + line_length]
            token_length = struct.unpack("<I", line[:INT_SIZE])[0]
            word = line[INT_SIZE : INT_SIZE + token_length].decode("utf-8")
            vector = np.frombuffer(line[max_length + INT_SIZE :], dtype=np.float32)
            yield word, vector


@file_or_name
def read_glove(f: Union[str, TextIO]) -> Tuple[Vocab, Vectors]:
    """Read vectors from a glove file.

    The GloVe format is a pure text format. Each (word, vector) pair is represented
    by a single line in the file. The line starts with the word, a space, and then
    the float32 text representations of the elements in the vector associated with
    that word. Each of these vector elements are also separated with a space.

    The main vectors distributed in this format are the `GloVe`_ vectors
    `(Pennington, et. al., 2014)`_

    .. _GloVe: https://nlp.stanford.edu/projects/glove/
    .. _(Pennington, et. al., 2014): https://www.aclweb.org/anthology/D14-1162/

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    return _read(f, read_glove_lines)


@file_or_name
def read_glove_with_vocab(
    f: Union[str, TextIO],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray] = uniform_initializer(0.25),
    keep_extra: bool = False,
) -> Tuple[Vocab, Vectors]:
    """Read vectors from a glove file subject to user vocabulary constraints.

    See :py:func:`~word_vectors.read.read_glove` for a description of the file format
    and common pre-train embeddings that use this format.

    When provided a vocabulary this function will not reorder it. If you pass in that
    the word ``dog`` is index ``12`` then in the resulting vocabulary it will still
    be index ``12``.

    When collecting extra vocabulary (words that are in the pre-trained embeddings but
    not in the user vocab) these will all be at the end of the vocabulary. Again the
    indices of user provided words will not change.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from.
        user_vocab: A specific vocabulary the user wants to extract form the pre-trained
            embeddings.
        initializer: A function that takes the vector size and generates a new vector.
            this is used to generate a representation for a word in the user vocab that
            is not in the pre-train embeddings.
        keep_extra: Should you also include vectors that are in the pre-trained embedding
            but not in the user provided vocab?

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if keep_extra:
        return _read_with_vocab_extra(f, read_glove_lines, user_vocab, initializer)
    return _read_with_vocab(f, read_glove_lines, user_vocab, initializer)


@file_or_name
def read_w2v_text(f: Union[str, TextIO]) -> Tuple[Vocab, Vectors]:
    """Read vectors from a text based w2v file.

    One of two different vector serialization formats introduced in the
    `word2vec software`_ `(Mikolov, et. al., 2013)`_.

    The word2vec text format is a pure text format. The first line is two
    integers, represented as text and separated by a space, that specify the
    number of types in the vocabulary and the size of the word vectors
    respectively. Each following line represents a (word, vector) pair. The
    line stars with the word, a space, and then the float 32 text representations
    of the elements in the vector associated with that word. Each of these vector
    elements are also separated with a space.

    One can see that that this is actually the same as the
    :py:func:`GloVe <word_vectors.read.read_glove>` format except that in GloVe
    they removed the header line.

    The main embeddings distributed in this format are
    `FastText`_ `(Bojanowski, et. al., 2017)`_ and
    `NumberBatch`_ `(Speer, et. al., 2017)`_


    .. _word2vec software: https://code.google.com/archive/p/word2vec/
    .. _(Mikolov, et. al., 2013): https://papers.nips.cc/paper/5021-distributed-representations-of-words-and-phrases-and-their-compositionality
    .. _FastText: https://fasttext.cc/
    .. _(Bojanowski, et. al., 2017): https://www.aclweb.org/anthology/Q17-1010/
    .. _NumberBatch: https://github.com/commonsense/conceptnet-numberbatch
    .. _(Speer, et. al., 2017): https://aaai.org/ocs/index.php/AAAI/AAAI17/paper/view/14972

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    return _read(f, read_w2v_text_lines)


@file_or_name
def read_w2v_text_with_vocab(
    f: Union[str, TextIO],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray] = uniform_initializer(0.25),
    keep_extra: bool = False,
) -> Tuple[Vocab, Vectors]:
    """Read vectors from a Word2Vec text file subject to user vocabulary constraints.

    See :py:func:`~word_vectors.read.read_w2v_text` for a description of the file format
    and common pre-train embeddings that use this format.

    When provided a vocabulary this function will not reorder it. If you pass in that
    the word ``dog`` is index ``12`` then in the resulting vocabulary it will still
    be index ``12``.

    When collecting extra vocabulary (words that are in the pre-trained embeddings but
    not in the user vocab) these will all be at the end of the vocabulary. Again the
    indices of user provided words will not change.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from.
        user_vocab: A specific vocabulary the user wants to extract form the pre-trained
            embeddings.
        initializer: A function that takes the vector size and generates a new vector.
            this is used to generate a representation for a word in the user vocab that
            is not in the pre-train embeddings.
        keep_extra: Should you also include vectors that are in the pre-trained embedding
            but not in the user provided vocab?

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if keep_extra:
        return _read_with_vocab_extra(f, read_w2v_text_lines, user_vocab, initializer)
    return _read_with_vocab(f, read_w2v_text_lines, user_vocab, initializer)


@file_or_name(f="rb")
def read_w2v(f: Union[str, BinaryIO]) -> Tuple[Vocab, Vectors]:
    """Read vectors from a word2vec file.

    One of two different vector serialization formats introduced in the
    `word2vec software`_ `(Mikolov, et. al., 2013)`_.

    The word2vec binary format is a mix of textual an binary representations.
    The first line is two integers (as text, separated be a space) representing
    the number of types in the vocabulary and the size of the word vectors
    respectively. (word, vector) pairs follow. The word is represented as text
    and a space. After the space each element of a vector is represented as a
    binary float32.

    The most well-known pre-trained embeddings distributed in this format are
    the `GoogleNews`_ vectors.

    Note:

        There is no formal definition of this file format, the only definitive
        reference on it is the original implementation in the `word2vec software`_

        Due to the lack of a definition (an no special handling of it in the code)
        there is no explicit statements about the endianness of the binary representations.
        Most code just uses the ``numpy.from_buffer`` and that seems to work now that
        most people have little-endian machines. However due to the lack of explicit
        direction on this encoding I would advise caution when loading vectors that
        were trained on big-endian hardware.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    .. _word2vec software: https://code.google.com/archive/p/word2vec/
    .. _(Mikolov, et. al., 2013): https://papers.nips.cc/paper/5021-distributed-representations-of-words-and-phrases-and-their-compositionality
    .. _GoogleNews: https://drive.google.com/file/d/0B7XkCwpI5KDYNlNUTTlSS21pQmM/edit

    Args:
        f: The file to read from

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    return _read(f, read_w2v_lines)


@file_or_name(f="rb")
def read_w2v_with_vocab(
    f: Union[str, BinaryIO],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray] = uniform_initializer(0.25),
    keep_extra: bool = False,
) -> Tuple[Vocab, Vectors]:
    """Read vectors from a Word2Vec file subject to user vocabulary constraints.

    See :py:func:`~word_vectors.read.read_w2v` for a description of the file format
    and common pre-train embeddings that use this format.

    When provided a vocabulary this function will not reorder it. If you pass in that
    the word ``dog`` is index ``12`` then in the resulting vocabulary it will still
    be index ``12``.

    When collecting extra vocabulary (words that are in the pre-trained embeddings but
    not in the user vocab) these will all be at the end of the vocabulary. Again the
    indices of user provided words will not change.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from.
        user_vocab: A specific vocabulary the user wants to extract form the pre-trained
            embeddings.
        initializer: A function that takes the vector size and generates a new vector.
            this is used to generate a representation for a word in the user vocab that
            is not in the pre-train embeddings.
        keep_extra: Should you also include vectors that are in the pre-trained embedding
            but not in the user provided vocab?

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if keep_extra:
        return _read_with_vocab_extra(f, read_w2v_lines, user_vocab, initializer)
    return _read_with_vocab(f, read_w2v_lines, user_vocab, initializer)


@file_or_name(f="rb")
def read_dense(f: Union[str, BinaryIO]) -> Tuple[Vocab, Vectors]:
    """Read vectors from a dense file.

    This is our fully binary vector format.

    The first line is a header for the dense format and it is a
    4-tuple. The elements of
    this tuple are: A magic number, the size of the vocabulary, the
    size of the vectors, and the length of the longest word in the
    vocabulary (this refers to the lengths of words when represented
    as ``utf-8`` bytes rather than as Unicode codepoints). These
    numbers are represented as little-endian unsigned long longs that
    have a size of 8 bytes.

    Following the header there are (length, word, vector) tuples. The
    length is the length of this particular word encoded as a
    little-endian unsigned integer. The word is stored as ``utf-8``
    bytes. The trick is that words are padded out to be a consistent
    length (this length is the length of the longest word in the
    vocabulary) but we keep track the length of this word to make reading
    the word from this padded section more efficient. After the word
    the vector is stored where each element is a little-endian
    float32 (4 bytes).

    The consistent word lengths lets us calculate the offset to any
    word quickly without having to iterate over the characters to
    find the space as in the word2vec binary format. Finding the
    word at index ``i`` can be done with some offset math.
    ``offset for i = header length + i * (max length + vector size + INT_SIZE)``

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from

    Returns:
        The vocab and vectors.
    """
    return _read(f, read_dense_lines)


@file_or_name(f="rb")
def read_dense_with_vocab(
    f: Union[str, BinaryIO],
    user_vocab: Vocab,
    initializer: Callable[[int], np.ndarray] = uniform_initializer(0.25),
    keep_extra: bool = False,
) -> Tuple[Vocab, Vectors]:
    """Read vectors from a Dense file subject to user vocabulary constraints.

    See :py:func:`~word_vectors.read.read_dense` for a description of the file format
    and common pre-train embeddings that use this format.

    When provided a vocabulary this function will not reorder it. If you pass in that
    the word ``dog`` is index ``12`` then in the resulting vocabulary it will still
    be index ``12``.

    When collecting extra vocabulary (words that are in the pre-trained embeddings but
    not in the user vocab) these will all be at the end of the vocabulary. Again the
    indices of user provided words will not change.

    Note:
       In the case of duplicated words in the saved vectors we use the index
       and associated vector from the first occurrence of the word.

    Args:
        f: The file to read from.
        user_vocab: A specific vocabulary the user wants to extract form the pre-trained
            embeddings.
        initializer: A function that takes the vector size and generates a new vector.
            this is used to generate a representation for a word in the user vocab that
            is not in the pre-train embeddings.
        keep_extra: Should you also include vectors that are in the pre-trained embedding
            but not in the user provided vocab?

    Returns:
        The vocab and vectors. The vocab is a mapping from word to integer and
        vectors are a numpy array of shape ``[vocab size, vector size]``. The
        vocab gives the index offset into the vector matrix for some word.
    """
    if keep_extra:
        return _read_with_vocab_extra(f, read_dense_lines, user_vocab, initializer)
    return _read_with_vocab(f, read_dense_lines, user_vocab, initializer)


@file_or_name(f="rb")
def sniff(f: Union[str, TextIO], buf_size: int = 1024) -> FileType:
    """Figure out what kind of vector file it is.

    Args:
        f: The file we are sniffing.
        buf_size: How many bytes to read in when sniffing the file.

    Returns:
        The guessed file type.
    """
    with bookmark(f):
        b = f.read(buf_size)
    # Because we support reading from an already open file we can't just start applying
    # the bytes based regexs to the file, if the file was not opened in binary mode use
    # the normal unicode string regexs
    if "b" not in f.mode:
        if GLOVE_TEXT.match(b):
            return FileType.GLOVE
        if W2V_TEXT.match(b):
            return FileType.W2V_TEXT
    else:
        if is_binary(f):
            if W2V_BIN.match(b):
                return FileType.W2V
            if verify_dense(b):
                return FileType.DENSE
        else:
            if W2V_BIN.match(b):
                return FileType.W2V_TEXT
            if GLOVE_BIN.match(b):
                return FileType.GLOVE
    raise ValueError(f"Could not determine file format for {f.name}")


def read_dense_header(buf: bytes) -> Tuple[int, int, int]:
    """Read the header from the dense file.

    The header for the dense format is a 4-tuple. The elements of
    this tuple are: A magic number, the size of the vocabulary, the
    size of the vectors, and the length of the longest word in the
    vocabulary (this length when represented as ``utf-8`` bytes
    rather than as Unicode codepoints). These numbers are represented
    as little-endian unsigned long longs that have a size of 8 bytes.

    Note:
        The magic number if used to make sure this is can actual
        file and not just trying to extract word vectors from a
        random binary file. The Magic Number is ``2283``.

    Args:
        buf: The beginning of the file we are reading the header from.

    Returns:
        The vocab size, the vector size, and the maximum length of any of the words

    Raises:
        ValueError: If the magic number doesn't match.
    """
    magic, vocab, dim, length = struct.unpack("<QQQQ", buf[: LONG_SIZE * DENSE_HEADER])
    if magic != DENSE_MAGIC_NUMBER:
        raise ValueError(f"Magic Number read does not match expected. Expected: `{DENSE_MAGIC_NUMBER}` Got: `{magic}`")
    return vocab, dim, length


def verify_dense(buf: bytes) -> bool:
    """Check if a file is in the dense format by comparing the magic number.

    Args:
        buf: The beginning of the file we are trying to determine if the it
        is a Dense formatted file.

    Returns:
        True if the magic number matched, False otherwise.
    """
    try:
        read_dense_header(buf)
        return True
    except ValueError:
        return False
